import re
from pathlib import Path

from labels.model.file import Coordinates, DependencyType, Location, LocationReadCloser, Scope
from labels.model.package import Language, Package, PackageType
from labels.parsers.cataloger.ruby.parse_gemfile import (
    create_dependency,
    format_requirements,
    parse_gemfile,
    parse_line,
)
from labels.testing.utils.helpers import get_test_data_path
from labels.utils.file import new_location


def test_parse_gem_file_lock() -> None:
    test_data_path = get_test_data_path("dependencies/ruby/Gemfile")
    expected_packages = [
        Package(
            name="minitest",
            version=">=5.15.0 <5.16",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=7,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/minitest@%3E%3D5.15.0%20%3C5.16",
        ),
        Package(
            name="minitest",
            version=">=5.15.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=9,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/minitest@%3E%3D5.15.0",
        ),
        Package(
            name="rake",
            version=">=11.1",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=13,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/rake@%3E%3D11.1",
        ),
        Package(
            name="sprockets-rails",
            version=">=2.0.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=15,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/sprockets-rails@%3E%3D2.0.0",
        ),
        Package(
            name="propshaft",
            version=">=0.1.7",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=16,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/propshaft@%3E%3D0.1.7",
        ),
        Package(
            name="capybara",
            version=">=3.26",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=17,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/capybara@%3E%3D3.26",
        ),
        Package(
            name="selenium-webdriver",
            version=">=4.0.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=18,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/selenium-webdriver@%3E%3D4.0.0",
        ),
        Package(
            name="rack-cache",
            version="^1.2",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=20,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/rack-cache@%5E1.2",
        ),
        Package(
            name="puma",
            version="=5.1.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=27,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/puma@%3D5.1.0",
        ),
        Package(
            name="rdoc",
            version="3.11",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=28,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/rdoc@3.11",
        ),
        Package(
            name="bcrypt",
            version="~3.1.11",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=32,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/bcrypt@~3.1.11",
        ),
        Package(
            name="terser",
            version=">=1.1.4",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=36,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/terser@%3E%3D1.1.4",
        ),
        Package(
            name="json",
            version=">=2.0.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=39,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/json@%3E%3D2.0.0",
        ),
        Package(
            name="rubocop",
            version=">=1.25.1",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=42,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/rubocop@%3E%3D1.25.1",
        ),
        Package(
            name="sdoc",
            version=">=2.4.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=50,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/sdoc@%3E%3D2.4.0",
        ),
        Package(
            name="redcarpet",
            version="~3.2.3",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=51,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/redcarpet@~3.2.3",
        ),
        Package(
            name="w3c_validators",
            version="~1.3.6",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=52,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/w3c_validators@~1.3.6",
        ),
        Package(
            name="rubyzip",
            version="^2.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=54,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/rubyzip@%5E2.0",
        ),
        Package(
            name="dalli",
            version=">=3.0.1",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=58,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/dalli@%3E%3D3.0.1",
        ),
        Package(
            name="listen",
            version="^3.3",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=59,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/listen@%5E3.3",
        ),
        Package(
            name="bootsnap",
            version=">=1.4.4",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=65,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/bootsnap@%3E%3D1.4.4",
        ),
        Package(
            name="queue_classic",
            version=">=4.0.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=75,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/queue_classic@%3E%3D4.0.0",
        ),
        Package(
            name="redis",
            version=">=4.0.1",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=85,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/redis@%3E%3D4.0.1",
        ),
        Package(
            name="google-cloud-storage",
            version="^1.11",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=93,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/google-cloud-storage@%5E1.11",
        ),
        Package(
            name="azure-storage-blob",
            version="^2.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=94,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/azure-storage-blob@%5E2.0",
        ),
        Package(
            name="image_processing",
            version="^1.2",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=96,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/image_processing@%5E1.2",
        ),
        Package(
            name="minitest-bisect",
            version="1.0.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=118,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/minitest-bisect@1.0.0",
        ),
        Package(
            name="stackprof",
            version="=1.4.2",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=123,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/stackprof@%3D1.4.2",
        ),
        Package(
            name="debug",
            version=">=1.1.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=124,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/debug@%3E%3D1.1.0",
        ),
        Package(
            name="nokogiri",
            version="=1.8.1",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=131,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/nokogiri@%3D1.8.1",
        ),
        Package(
            name="devise",
            version="=1.4.2",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=132,
                    ),
                    scope=Scope.DEV,
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=True,
            metadata=None,
            p_url="pkg:gem/devise@%3D1.4.2",
        ),
        Package(
            name="rubocop",
            version="=1.35.1",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=133,
                    ),
                    scope=Scope.DEV,
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=True,
            metadata=None,
            p_url="pkg:gem/rubocop@%3D1.35.1",
        ),
        Package(
            name="rails",
            version="=7.0.4",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=134,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/rails@%3D7.0.4",
        ),
        Package(
            name="racc",
            version=">=1.4.6",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=137,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/racc@%3E%3D1.4.6",
        ),
        Package(
            name="sqlite3",
            version="^1.4",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=140,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/sqlite3@%5E1.4",
        ),
        Package(
            name="pg",
            version="^1.3",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=143,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/pg@%5E1.3",
        ),
        Package(
            name="mysql2",
            version="^0.5",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=144,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/mysql2@%5E0.5",
        ),
        Package(
            name="pry",
            version="=1.0.1",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=145,
                    ),
                    dependency_type=DependencyType.DIRECT,
                    access_path=test_data_path,
                    annotations={},
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/pry@%3D1.0.1",
        ),
        Package(
            name="activerecord-jdbcsqlite3-adapter",
            version=">=1.3.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=157,
                    ),
                    access_path=test_data_path,
                    annotations={},
                    dependency_type=DependencyType.DIRECT,
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/activerecord-jdbcsqlite3-adapter@%3E%3D1.3.0",
        ),
        Package(
            name="activerecord-jdbcmysql-adapter",
            version=">=1.3.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=159,
                    ),
                    access_path=test_data_path,
                    annotations={},
                    dependency_type=DependencyType.DIRECT,
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/activerecord-jdbcmysql-adapter@%3E%3D1.3.0",
        ),
        Package(
            name="activerecord-jdbcpostgresql-adapter",
            version=">=1.3.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=160,
                    ),
                    access_path=test_data_path,
                    annotations={},
                    dependency_type=DependencyType.DIRECT,
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/activerecord-jdbcpostgresql-adapter@%3E%3D1.3.0",
        ),
        Package(
            name="ruby-oci8",
            version="^2.2",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=168,
                    ),
                    access_path=test_data_path,
                    annotations={},
                    dependency_type=DependencyType.DIRECT,
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/ruby-oci8@%5E2.2",
        ),
        Package(
            name="wdm",
            version=">=0.1.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=174,
                    ),
                    access_path=test_data_path,
                    annotations={},
                    dependency_type=DependencyType.DIRECT,
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/wdm@%3E%3D0.1.0",
        ),
        Package(
            name="mini_magick",
            version="4.9.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=175,
                    ),
                    access_path=test_data_path,
                    annotations={},
                    dependency_type=DependencyType.DIRECT,
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/mini_magick@4.9.0",
        ),
        Package(
            name="rails-html-sanitizer",
            version="^1.2.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=176,
                    ),
                    access_path=test_data_path,
                    annotations={},
                    dependency_type=DependencyType.DIRECT,
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/rails-html-sanitizer@%5E1.2.0",
        ),
        Package(
            name="error_highlight",
            version=">=0.4.0",
            language=Language.RUBY,
            licenses=[],
            locations=[
                Location(
                    coordinates=Coordinates(
                        real_path=test_data_path,
                        file_system_id=None,
                        line=182,
                    ),
                    access_path=test_data_path,
                    annotations={},
                    dependency_type=DependencyType.DIRECT,
                ),
            ],
            type=PackageType.GemPkg,
            advisories=None,
            dependencies=None,
            found_by=None,
            health_metadata=None,
            is_dev=False,
            metadata=None,
            p_url="pkg:gem/error_highlight@%3E%3D0.4.0",
        ),
    ]
    with Path(test_data_path).open(encoding="utf-8") as reader:
        pkgs, _ = parse_gemfile(
            None,
            None,
            LocationReadCloser(location=new_location(test_data_path), read_closer=reader),
        )
        for pkg in pkgs:
            pkg.health_metadata = None
            pkg.licenses = []
        assert pkgs == expected_packages


def test_format_requirements_with_greater_equal() -> None:
    requirements = ["~>1.0.0", ">=2.0.0"]
    result = format_requirements(requirements)
    assert result == "^2.0.0"


def test_format_requirements_with_greater_equal_and_not_equal() -> None:
    requirements = ["~>1.0.0", ">=2.0.0", "!=3.0.0"]
    result = format_requirements(requirements)
    assert result == ""


def test_format_requirements_with_not_equal() -> None:
    requirements = ["~>1.0.0", "2.0.0"]
    result = format_requirements(requirements)
    assert result == ""


def test_format_requirements_with_not_equal_operator() -> None:
    requirements = ["~>1.0.0", "!=2.0.0"]
    result = format_requirements(requirements)
    assert result == ">=~>1.0.0 <2.0.0  || >2.0.0"


def test_format_requirements_unhandled_case() -> None:
    requirements = ["random", "???"]
    result = format_requirements(requirements)
    assert result == "random ???"


def test_parse_line_without_comma() -> None:
    in_line = "gem_name 1.0.0"
    product, version = parse_line(in_line, gem_file=False)
    assert product == "gem_name"
    assert version == "1.0.0"


def test_parse_line_with_comma() -> None:
    in_line = "gem_name 1.0.0, 2.0.0"
    product, version = parse_line(in_line, gem_file=False)
    assert product == "gem_name"
    assert version == "1.0.0 2.0.0"


def test_create_dependency_without_coordinates() -> None:
    location = Location(
        coordinates=None,
        dependency_type=DependencyType.DIRECT,
        access_path="test_path",
        annotations={},
    )

    mock_match = re.match(
        r'^\s*(?P<gem>gem ".*?",?( "[><~=]{0,2}\s?[\d\.]+",?){0,2})',
        'gem "test_gem", "1.2.3"',
    )
    assert mock_match is not None

    result = create_dependency(
        line_number=1,
        location=location,
        matched=mock_match,
        is_dev=False,
    )

    assert result is not None
    assert result.name == "test_gem"
    assert result.locations[0].coordinates is None
    assert result.locations[0].dependency_type == DependencyType.DIRECT
